#include "BrianExpert.h"
#include <iostream>
#include <stc.h>
#include <msg.h>
#include <MrtMisc.h>

#include "smdebug.h"
#include "Configurator.h"

extern void brianlex_memory(const char* src, BrianExpert* brian);

//#define BRIAN_DEBUG 1 //*wiim*
//#define EXPERTS_DEBUG

const string BrianExpert::msg_always = "Always";

BrianExpert::BrianExpert(ModuleDispatch* kernel,long period) :
  StringModuleMember(kernel,period,"brian"){

	//m_bTurning = false;

#ifdef EXPERTS_DEBUG
  cout<<"\n---------------> Starting Brian" << endl;
#endif

  mLogger.open("log/mylog.txt", ios::out );

  TheBrian = new MrBrian(
		(char*)Configurator::brian_ctof.c_str(),	  				//"./config/brian/spykee/ctof.txt",
  	 	(char*)Configurator::brian_ctof_shapes.c_str(), 			//"./config/brian/spykee/shape_ctof.txt",
  	 	(char*)Configurator::brian_predicate.c_str(),				//"./config/brian/spykee/Predicate.ini",
  	 	(char*)Configurator::brian_predicate_actions.c_str(),		//"./config/brian/spykee/PredicateActions.ini",
  	 	(char*)Configurator::brian_cando.c_str(),					//"./config/brian/spykee/Cando.ini",
  	 	(char*)Configurator::brian_behaviour.c_str(),				//"./config/brian/spykee/behaviour.txt",
  	 	(char*)Configurator::brian_want.c_str(),					//"./config/brian/spykee/want.ini",
  	 	(char*)Configurator::brian_ftoc.c_str(),					//"./config/brian/spykee/s_ftoc.txt",
  	 	(char*)Configurator::brian_ftoc_shapes.c_str()				//"./config/brian/spykee/s_shape.txt"
  	 	);
}

BrianExpert::~BrianExpert(){
  logfile.close();
  mLogger.close();
  delete TheBrian;
}


void BrianExpert::RunInit(){
	mLogger << "*** BRIAN INIT ***" << endl;

	AddMessageRequest(MSG_FROM_SONAR,LOCAL);
	AddMessageRequest(MSG_FROM_WIIMOTE, LOCAL );
	AddMessageRequest(MSG_FROM_VISION, LOCAL );

	logfile.open(BRIANFULLLOG);
	brianmsglog.open(BRIANMSGLOG, ios::out|ios::app );

	AddValueToMap( "GameStatus", STATUS_SEARCHING, 1 );		// SEARCHING
	AddValueToMap( "Always", 1, 1 );						// TRUE -- utility
	AddValueToMap( "RotSearch", 0, 1 );						// LEFT

	mMotorState = MOTOR_NORMAL;
	mEscapeTimeout = Configurator::escape_timeout;
	mOldClock = false;
	settimer(0);
}

void BrianExpert::RunClose(){
	mLogger << "\n*** BRIAN CLOSE ***" << endl;
	logfile.close();
	brianmsglog.close();
}

void BrianExpert::RunDuty(){
	bool bReceived = false;
	bool bStop = false;

	unsigned char *buffer = NULL;

	if ( ReceiveLastMessage( MSG_FROM_SONAR, buffer, false ) > 0 ){
		brianmsglog << '\n' << buffer << endl;
		brianlex_memory( (const char*)buffer, this );
		bReceived = true;
		buffer = NULL;
	}

	if ( ReceiveLastMessage( MSG_FROM_VISION, buffer, false ) > 0 ){
		brianmsglog << '\n' << buffer << endl;
		brianlex_memory( (const char*)buffer, this );
		bReceived = true;
		buffer = NULL;
	}

	if ( ReceiveLastMessage( MSG_FROM_WIIMOTE, buffer, false ) > 0 ){
		brianmsglog << '\n' << buffer << endl;
		brianlex_memory( (const char*)buffer, this );
		bReceived = true;
		buffer = NULL;
	}

	// Se non ho inviato il segnale di gioco, non faccio nulla
	if ( mCL.find("GameStart") == mCL.end() || (mCL.find("GameStart") != mCL.end() && mCL.find("GameStart")->second.first == 0) ){
		// resetto mappa e brian
		mCL.clear();

		// workaroud -> brian problems!
		TheBrian->run();
	 	TheBrian->flush();

		// Imposto i valori iniziali
		AddValueToMap( "GameStatus", STATUS_SEARCHING, 1 );		// SEARCHING
		AddValueToMap( "Always", 1, 1 );						// TRUE -- utility
		AddValueToMap( "RotSearch", 0, 1 );						// LEFT
		AddValueToMap( "GameStart", 0, 1 );

		mMotorState = MOTOR_NORMAL;
		mEscapeTimeout = Configurator::escape_timeout;
		mOldClock = false;

		return;
	}

//	cout << "RUN" << endl;

	BrianMap::iterator clkIt = mCL.find( "Clock" );				// clock -- settato da WiimExpert
	BrianMap::iterator statusIt = mCL.find("GameStatus");		// Stato del gioco
	BrianMap::iterator irIt = mCL.find("IRCenter");				// Puntato ?
	BrianMap::iterator vIt = mCL.find("VisionPlayerDetected");  // Player in vista

	/* Se il robot sta scappando da piu di 10 secondi e non viene nuovamente puntato e non ha trovato un nascondiglio,
	 * ritorna in stato di ricerca.
	 * Se viene nuovamente puntato il timer si resetta.
	 */

	if ( irIt == mCL.end() || vIt == mCL.end() ){ // Se non è puntato e non vede il giocatore
		if ( clkIt != mCL.end() && statusIt->second.first == STATUS_ESCAPING ){ // Se sta scappando
			if ( ((bool)clkIt->second.first) != mOldClock ){			// ed è passato un altro secondo
				mOldClock = (bool)clkIt->second.first;					// imposto il clock attuale
				mEscapeTimeout--;										// decremento di un altro secondo
				cout << "Escape reset in: " << mEscapeTimeout << endl;	// debug
			}
		}
	} else {	// Se il robot è puntato o vede il giocatore
		mEscapeTimeout = Configurator::escape_timeout;	//continua a scappare
	}

	/*
	 * Timeout fuga: se dopo 10 (default) secondi il robot non viene nuovamente puntato o non trova un nascondiglio
	 * torno in modalità di ricerca
	 */

	if ( mEscapeTimeout <= 0 ){
		mEscapeTimeout = Configurator::escape_timeout;
		AddValueToMap( "GameStatus", STATUS_SEARCHING, 1 );
		mLogger << "INFO: EscapeReset" << endl;
	}


	// Copio i valori della mappa nella crisp_data_list
	for ( BrianMap::iterator it = mCL.begin(); it != mCL.end(); it++ )
		SetValue( it->first.c_str(), it->second.first, it->second.second ); //name - value - rel = 1

	// Mando lo stato del gioco al wiimote expert per visualizzarlo a console / gui
	NewMessage(MSG_FROM_BRIAN,"info");
	mCurrentMessage << "<D>GameStatus " << mCL.find("GameStatus")->second.first << "</D>";
	CloseCurrentMessage();
	SendAllStringMessages();

	// Eseguo Mr. Brian
	TheBrian->run();

	// Ottengo le variabili crisp in uscita
	command_list *cmdList = TheBrian->getFuzzy()->get_command_singleton_list();

	/*
	 * Retroazione dello stato di gioco ( SEARCH, ESCAPE, HIDE )
	 */

	command *cmdStatus = cmdList->get_command("GameStatus");

	if ( cmdStatus ){
		mGameStatus = cmdStatus->get_set_point();
		AddValueToMap( "GameStatus", (int)mGameStatus );
	}

	/*
	 * Retoazione della direzione di ricerca
	 */

	command *cmdRotSearch = cmdList->get_command("RotSearch");

	if ( cmdRotSearch )
		AddValueToMap( "RotSearch", cmdRotSearch->get_set_point() );

	/*
	 * Gestion della rotazione a sx/dx di 90° gradi
	 */

	BrianMap::iterator gsCl = mCL.find("GameStatus");

	// Se il robot si sta nascondendo
	if ( gsCl != mCL.end() && gsCl->second.first == STATUS_HIDING ){
		command *cmdMotor = cmdList->get_command( MOTOR_COMMAND );
		predicate *irExists = TheBrian->get_predicate_list()->get("IRExist");

		// Se il robot è puntato
		if ( irExists ){
			mMotorState = MOTOR_NORMAL;
			mEscapeTimeout = Configurator::escape_timeout;
			RemoveMapValue("AutoTurn");	 // Distattiva l'attivazione delle regole Hides
			RemoveMapValue("GoForward"); // Come sopra

			//Riporta il robot nello stato di fuga
			AddValueToMap( "GameStatus", STATUS_ESCAPING );
			cout << "End hide -> ESCAPE" << endl;

			goto out;
		}

		// Il robot vuole nascondersi a destra
		if ( cmdMotor && cmdMotor->get_set_point() == SPYKEE_TURN_RIGHT ){
			// Qui inizia la rotazione, MOTOR_NORMAL indica che il controllo dei motori è affidato a
			// MotorExpert
			if ( mMotorState==MOTOR_NORMAL ){
				// Per tenere attiva la regola HideRight
				AddValueToMap("AutoTurn", 1, 1);
				mMotorState = SPYKEE_TURN_RIGHT;
				//Inizializzo il timer per caloclare 90°
				settimer(0);
				cout << "Start turning right" << endl;
			}

			/* Quando il timer dice che il tempo è passato, il robot smette di girare
			 * e avanza per nascondersi
			 */

			if ( elapsed(0) > Configurator::turn_time && (mMotorState==SPYKEE_TURN_RIGHT) ){
				mMotorState = SPYKEE_AUTO_FORWARD;
				AddValueToMap("GoForward", 1, 1);
				cout << "End turning right -> GoForward" << endl;
				settimer(0);
			}

		} else if ( cmdMotor && cmdMotor->get_set_point() == SPYKEE_TURN_LEFT ){
			/*
			 * Come sopra, ma per girare a sinsitra
			 */
			if ( mMotorState==MOTOR_NORMAL ){
				//Per tenere attiva la regola HideLeft
				AddValueToMap("AutoTurn", 0, 1);
				mMotorState = SPYKEE_TURN_LEFT;
				settimer(0);
				cout << "Start turning left" << endl;
			}

			if ( elapsed(0) > Configurator::turn_time && (mMotorState==SPYKEE_TURN_LEFT) ){
				mMotorState = SPYKEE_AUTO_FORWARD;
				AddValueToMap("GoForward", 1, 1);
				cout << "End turning left -> GoForward" << endl;
				settimer(0);
			}
		}

	} else if (gsCl != mCL.end() && gsCl->second.first == STATUS_SEARCHING ){
		command *cmdMotor = cmdList->get_command( MOTOR_COMMAND );
		/*
		 * Il robot si è nascosto
		 */
		if ( (cmdMotor && cmdMotor->get_set_point() == 6 ) && (mMotorState == SPYKEE_AUTO_FORWARD) ){
			/*
			 * Restituisco il controllo dei motori al MotorExpert e torno in modalità di ricerca
			 */
			mMotorState = MOTOR_NORMAL;
			RemoveMapValue("AutoTurn");
			RemoveMapValue("GoForward");
			cout << "End hide -> Search" << endl;

			AddValueToMap("GameStatus", STATUS_SEARCHING, 1.0 );

			/*
			 * Se non si è nascosto bene ricomincia a cercare un nascondiglia
			 */
			predicate *pRight = TheBrian->get_predicate_list()->get("RightBusy");
			predicate *pLeft = TheBrian->get_predicate_list()->get("LeftBusy");

			/*if ( pRight == NULL && pLeft == NULL ){
				AddValueToMap("GameStatus", STATUS_ESCAPING, 1.0 );
				this->mEscapeTimeout = Configurator::escape_timeout;
				cout << "HIDE FAILED -> Escape again" << endl;
			}*/

		}
	}

	// Se ricevo qualcosa da qualche expert lo loggo
/*	if ( bReceived ){

	}
*/
	// Creazione file di log
	TheBrian->log();
	debug();

	// Se ho qualcosa da inviare lo mando ai motori
	if ( cmdList->size() > 0 ){
		NewMessage(MSG_TO_MOTION, COMMAND_DATA);

		for ( command_list::iterator it = cmdList->begin(); it != cmdList->end(); it++ )
			AddCommandData( it->first, it->second->get_set_point() );

		CloseCurrentMessage();
		SendAllStringMessages();
	}

out:
	// Pulisco la memoria di Mr. Brian
	TheBrian->flush();

	/* Cancello i valori che indicano se il robot è puntato,
	 * altrimenti una volta puntato rimarrebbe sempre in questo stato
	 */

	RemoveMapValue("IRCenter");
	RemoveMapValue("IRDist");
}

void BrianExpert::RemoveMapValue(string name){
	mCL.erase(name);
	mLogger << "MANUAL: " << name << " removed" << endl;
}

// Utility
void BrianExpert::ParseStringMessages(){
	unsigned char *buffer = NULL;

	if (ReceiveLastMessage(MSG_TO_BRIAN,buffer,true)>0){
		FILE* fp=fopen(BRIANMSGLOG,"a");
		fprintf(fp,"\n%s\n",buffer);
		fclose(fp);
		brianlex_memory((const char*)buffer,this);
	}
}

void BrianExpert::FillMessages(command_list * cl, action_list * al, predicate_list* linkcadl, weight_want_list* linkwwl, predicate_list* linkpdl){
	// create the message to return

	command_multimap::iterator c_it;
	command* c = 0;
	for (c_it = cl->begin(); c_it != cl->end(); c_it++){
		c = c_it->second;
		AddCommandData(c_it->first,c->get_set_point());
		cout << c_it->first << ": " << c->get_set_point() << ' ';
	}

	cout << endl;
}

void BrianExpert::SetValue(const char * name, float val, float rel){
	// search for the crisp_data whose naAddCommandData("RotSpeed", 0.0 );me is "name" and sets its value and reliability to "val" and "rel"
	crisp_data_list * cdl = (TheBrian->getFuzzy())->get_crisp_data_list();
	crisp_data_list::iterator i=cdl->find(name);
  
	if (i != cdl->end()){
		(*i).second->set_value(val);
		(*i).second->set_reliability(rel);
	} else {
	  cdl->add(new crisp_data(name,val,rel));
  }
}

void BrianExpert::AddBehaviorData(string name, float cando_value, float want_value){
	mCurrentMessage << "<B>" << name << " " << cando_value << " " << want_value << "</B>\n";
}

void BrianExpert::AddFuzzyData(string name, float value, float reliability ){
	mCurrentMessage << "<F>" << name << " " << value << " " << reliability << "</F>\n";
}


void BrianExpert::AddValueToMap( const char *name, float val, float rel ){
	mCL[name] = std::pair<float,float>( val, rel );
}

void BrianExpert::AddCommandData(string name, float value){
	if ( fabs(value) < 0.001 ) //*wiim* evito che vengano mandati valori in notazione scientifica
		value = 0;
	
	mCurrentMessage << "<C>" << name << ' ' << value << "</C>\n";
}

long BrianExpert::elapsed(int i){
	struct timeval now;
	long elapsed_usec = 0;

	gettimeofday(&now, NULL);

	if (now.tv_usec > 1000000) {
		now.tv_usec -= 1000000;
		++now.tv_sec;
	}

	if (now.tv_sec > last_render[i].tv_sec){
		elapsed_usec = ((now.tv_sec - last_render[i].tv_sec) * 1000000);
	}

	if (now.tv_usec > last_render[i].tv_usec)
		elapsed_usec += now.tv_usec - last_render[i].tv_usec;
	else
		elapsed_usec -= last_render[i].tv_usec - now.tv_usec;

	return elapsed_usec ;
}

void BrianExpert::settimer(int i){
	gettimeofday(&(last_render[i]),NULL);
}

void BrianExpert::debug(){
	struct timeval now;
	gettimeofday(&now,NULL);

	//ostringstream ostr;
	mLogger << "\n\n" << now.tv_sec << "_" << now.tv_usec << " Brian debug start **********************\n" << endl;
	mLogger.setf( ios::left, ios::adjustfield );
	//mLogger.setf( ios::fixed, ios::floatfield );

	crisp_data_list *cl = TheBrian->getFuzzy()->get_crisp_data_list();
	mLogger << "--------- Crisp Data - Length: " << cl->size() << "---------" << endl;
	for ( crisp_data_list::iterator it = cl->begin(); it != cl->end(); it++  ){
		mLogger << "CRISP: " << setw(20) << it->second->get_name()
				<< " = " << fixed << setw(4) << it->second->get_value() << "\t"
				<< setw(4) << it->second->get_reliability() << endl;
	}
	mLogger << "----- End Crisp Data -----\n" << endl;


	fuzzy_data_list *fl = TheBrian->getFuzzy()->get_fuzzy_data_list();
	mLogger << "--------- Fuzzy Data - Length: " << fl->size() << " ---------" << endl;
	for ( fuzzy_data_list::iterator it = fl->begin(); it != fl->end(); it++ ){
		mLogger << "FUZZY: " << setw(20) << it->second->get_name()
				<< " = " << setw(15) << it->second->get_label()
				<< setw(4) << it->second->get_membership_value() << endl;
	}
	mLogger <<  "--------- Fuzzy Data End -----\n" << endl;


	predicate_list *pl = TheBrian->get_predicate_list();
	mLogger << "--------- Predicates - Length: " << pl->size() << " ---------" << endl;
	for ( predicate_list::iterator it = pl->begin(); it != pl->end(); it++ ){
		mLogger << "PREDICATE: " << setw(20) << it->second->get_name()
				<< " = " << setw(4) << it->second->get_value()
				<< " rel = " << setw(4) << it->second->get_reliability() << endl;
	}
	mLogger << "----- End Predicates -----\n" << endl;


	predicate_list *canl = TheBrian->get_cando_list();
	mLogger << "--------- Cando - Length: " << canl->size() << " ---------" << endl;
	for ( predicate_list::iterator it = canl->begin(); it != canl->end(); it++ ){
		mLogger << "CANDO: " << setw(20) << it->second->get_name()
				<< " = " << setw(4) << it->second->get_value()
				<< " rel = " << setw(4) << it->second->get_reliability() << endl;
	}
	mLogger << "----- End Cando -----\n" << endl;


	proposed_action_list *pal = TheBrian->get_proposed_action_list();
	mLogger << "--------- Proposed Action - Length: " << pal->size() << " ---------" << endl;
	for ( proposed_action_list::iterator it = pal->begin(); it != pal->end(); it++ ){
		mLogger << "PAL: " << setw(20) << it->second->get_name() << " FROM " << setw(10) << it->second->get_behavior_name()
				<< " = " << setw(15) << it->second->get_label()
				<< " rel = " << setw(4) << it->second->get_membership_value() << endl;
	}
	mLogger << "----- End Proposed Action -----\n" << endl;


	weight_want_list *wl = TheBrian->get_weight_want_list();
	mLogger << "--------- Want - Length: " << wl->size() << " ---------" << endl;
	for ( weight_want_list::iterator it = wl->begin(); it != wl->end(); it++ ){
		mLogger << "WANT: " << setw(20) << it->second->get_name()
				<< " = " << setw(4) << it->second->get_value() << endl;
	}
	mLogger << "----- End Want -----\n" << endl;


	action_list *al = TheBrian->getFuzzy()->get_action_list();
	mLogger << "--------- Action List - Length: " << al->size() << " ---------" << endl;
	for ( action_list::iterator it = al->begin(); it != al->end(); it++ ){
		mLogger << "ACTION: " << setw(20) << it->second->get_name()
				<< " = " << setw(15) << it->second->get_label()
				<< " rel = " << setw(4) << it->second->get_membership_value();

		if ( it->second->get_membership_value() == 0 )
			mLogger << "\t" << "DISABLED";

		mLogger << endl;
	}
	mLogger << "----- End Action -----\n" << endl;

	command_list *cmdl = TheBrian->getFuzzy()->get_command_singleton_list();
	mLogger << "--------- Commnad List - Length: " << cmdl->size() << " ---------" << endl;
	for ( command_list::iterator it = cmdl->begin(); it != cmdl->end(); it++ ){
		mLogger << "COMMAND: " << setw(20) << it->second->get_label() << ' '
				<< setw(4) << it->second->get_set_point() << endl;
	}
	mLogger << "----- End Command -----\n" << endl;

	mLogger << "***** Brian debug end *****************************************************************\n" << endl;
}
