/*
 * Configurator.cpp
 *
 *  Created on: 31/mag/2010
 *      Author: andrea
 */

#include "Configurator.h"
#include <boost/program_options.hpp>

#include <fstream>
#include <iostream>

using namespace std;
namespace po = boost::program_options;

int Configurator::escape_timeout;
int Configurator::turn_time;

int Configurator::sonar_delay;
int Configurator::sonar_front;
int Configurator::sonar_back;
int Configurator::sonar_right;
int Configurator::sonar_left;
int Configurator::hide_depth;

string Configurator::sonar_source;

string Configurator::spykee_user;
string Configurator::spykee_pass;

bool Configurator::initAuto;
bool Configurator::simulation;

string Configurator::brian_ctof_shapes;
string Configurator::brian_ctof;
string Configurator::brian_predicate;
string Configurator::brian_predicate_actions;
string Configurator::brian_cando;
string Configurator::brian_want;
string Configurator::brian_behaviour;
string Configurator::brian_ftoc;
string Configurator::brian_ftoc_shapes;

bool Configurator::m_bInit = false;

Configurator::Configurator() {
}

Configurator::~Configurator() {
}


bool Configurator::init( const string &path, int argc, char **argv ){
	if ( !m_bInit ){
		try {
			po::options_description opt("ConfigFile");
			po::options_description opt2("CommandLine");

			opt.add_options()
				( "sonar-source", po::value<string>(&sonar_source)->default_value("/dev/ttyUSB0"), "Sonar serial port" )
				( "sonar-delay", po::value<int>(&sonar_delay)->default_value(1000000), "Sonar old delay" )
				( "sonar-front", po::value<int>(&sonar_front)->default_value(0), "Front sonar index")
				( "sonar-back", po::value<int>(&sonar_back)->default_value(0), "Back sonar index")
				( "sonar-right", po::value<int>(&sonar_right)->default_value(0), "Right sonar index")
				( "sonar-left", po::value<int>(&sonar_left)->default_value(0), "Left sonar index")
				( "hide-delay", po::value<int>(&turn_time)->default_value(900000), "Hide dealy")
				( "escape-timeout", po::value<int>(&escape_timeout)->default_value(10), "Escape timeout")
				( "hide-depth", po::value<int>(&hide_depth)->default_value(500), "Hide depth")

				( "spykee-user", po::value<string>(&spykee_user)->default_value("admin"), "Spykee username")
				( "spykee-pass", po::value<string>(&spykee_pass)->default_value("admin"), "Spykee password")

				( "auto", po::value<bool>(&initAuto)->default_value(false), "Start with game" )
				( "sim", po::value<bool>(&simulation)->default_value(false), "Spykee simulation" )

				( "brian-ctof", po::value<string>(&brian_ctof), "Crisp to fuzzy" )
				( "brian-ctof-shapes", po::value<string>(&brian_ctof_shapes), "Crisp to fuzzy shapes" )
				( "brian-cando", po::value<string>(&brian_cando), "Cando" )
				( "brian-want", po::value<string>(&brian_want), "Want" )
				( "brian-behaviour", po::value<string>(&brian_behaviour), "behaviour" )
				( "brian-predicate", po::value<string>(&brian_predicate), "Predicates" )
				( "brian-predicate-actions", po::value<string>(&brian_predicate_actions), "PredicateActions" )
				( "brian-ftoc", po::value<string>(&brian_ftoc), "Fuzzy to crisp" )
				( "brian-ftoc-shapes", po::value<string>(&brian_ftoc_shapes), "Fuzzy to crisp shapes" );


			opt2.add_options()
				( "auto", po::value<bool>(), "Game autostart" )
				( "sim", po::value<bool>(), "Spykee simulation" );

			po::variables_map vm_file;
			ifstream config_file(path.c_str());
			po::store( po::parse_config_file(config_file, opt), vm_file );
			config_file.close();
			notify(vm_file);

			po::variables_map vm_cmd;
			po::store( po::parse_command_line(argc, argv, opt2), vm_cmd );
			notify(vm_cmd);

			m_bInit = true;

			if ( vm_cmd.count("auto") == 1 )
				initAuto = vm_cmd["auto"].as<bool>();

			if ( vm_cmd.count("sim") == 1 )
				simulation = vm_cmd["sim"].as<bool>();


			cout << "AutoStart: " << boolalpha << initAuto << '\n'
				 << "Simulation: " << boolalpha << simulation << endl;
		} catch ( exception &e ){
			cout << e.what() << endl;
			return false;
		}
	}

	return true;
}
