#include "SonarExpert.h"

#include <iostream>
#include <deque>
#include <string>

#include "ReadSonar.h"
#include "Configurator.h"

using namespace std;
//using namespace log4cpp;

SonarExpert::SonarExpert(ModuleDispatch* kernel,std::string name,int period)
 	: StringModuleMember(kernel,period,name), readSonar(NULL)
 {
	 try {
		readSonar = new ReadSonar( Configurator::sonar_source, 1 );
		rLog.open( "log/rlog.txt" );
		lLog.open( "log/llog.txt" );
	} catch ( ReadSonarDeviceException &e ){
		printf( "\n%s\n", e.what() );
		readSonar = NULL;
	} catch (...){
		cout << "Big error: NO SONAR!!" << endl;
	}
}

SonarExpert::~SonarExpert(){
	if( readSonar )
		delete readSonar;

	lLog.close();
	rLog.close();
}


void SonarExpert::RunInit(){
	try {
		if ( readSonar )
			readSonar->sendRun();
		else
			cerr << "Sonar is null!!!" << endl;
		settimer(0);
	} catch (ReadSonarDeviceException &e) {
		printf( "ERRORE SONAR\n" );
	}
	
	AddMessageRequest( MSG_TO_SONAR, LOCAL );

	old_sonar_values[0] = old_sonar_values[1] = 10000;
	m_ledValue.cValue = 0xF;

	printf( "\nSonarExpert::RunInit()\n" );

	SetAllLeds(0);

	m_bBufferFull = m_bInit = false;
	//m_lBuffer.push_back(10000);
	//m_rBuffer.push_back(10000);
	settimer(0);
}

void SonarExpert::RunDuty() {
	int meas_progress=0;
	int line_readed=0;
	float sonarL, sonarR;

	if ( !readSonar ){
		return;
	}

	unsigned char *buffer = NULL;

	if ( ReceiveLastMessage( MSG_TO_SONAR, buffer, false ) > 0 ){
		float n = 0;
		int i = 0;

		for(i=0;buffer[i]!=NULL;i++);
    	char *buf = (char*)malloc(i+1);
    	memset( buf, 0 , i );
    	strcpy(buf,(char*)buffer);

    	if ( strtovalue( buf, "Red ", &n ) )
    		m_ledValue.sValue.red = (char)n;

		if ( strtovalue( buf, "Yellow ", &n ) )
			m_ledValue.sValue.yellow = (char)n;

		if ( strtovalue( buf, "Green ", &n ) )
			m_ledValue.sValue.green = n > 0.0 ? 1 : 0;

    	free(buf);

    	readSonar->sendStringCommand(&m_ledValue.cValue,1);
	}

	do {		
		if(readSonar->readData()==0){
			unsigned int n_line;
			n_line=readSonar->getLineToParseNum();
			line_readed+=n_line;
			
			for(unsigned int i=0;i<n_line;i++){	
				switch(readSonar->parseLine()){
					case ReadSonar::parse_err:
						//printf("error in parse method <%s>",readSonar->getParsedLine());
						meas_progress=0;
						break;
					case ReadSonar::parse_meas_b1:
						//FileLogger::write("found meas group 1:");
						//FileLogger::writeLine(readSonar->getParsedLine());
						if(meas_progress==0)
							meas_progress++;
						break;
					case ReadSonar::parse_meas_b2:
						//FileLogger::write("found meas group 2:");
						//FileLogger::writeLine(readSonar->getParsedLine());
						if(meas_progress==1)
							meas_progress++;
						break;
					case ReadSonar::parse_meas_b3:
						//FileLogger::write("found meas group 3:");
						//FileLogger::writeLine(readSonar->getParsedLine());
						if(meas_progress==2)
							meas_progress++;
						break;
/*					case ReadSonar::parse_meas_b4:
						FileLogger::write("found meas group 4:");
						FileLogger::writeLine(readSonar->getParsedLine());
						if(meas_progress==3)meas_progress++;
						break;*/
					case ReadSonar::parse_dbg:
						printf("found DBG line <%s>",readSonar->getParsedLine());
						meas_progress=0;
						break;
//					case ReadSonar::parse_ok:
//						printf("found OK line <%s>",readSonar->getParsedLine());
//						meas_progress=0;
//						break;
					case ReadSonar::parse_response_err:
						printf("found ERR line <%s>",readSonar->getParsedLine());
						meas_progress=0;
						break;
					default:
						printf("defensive programming... unrechable... mumble mumble\n");
						meas_progress=0;
						break;
				}

				//meas_progress = 3;

				if(meas_progress != 0){
					if ( readSonar->getMeasure(Configurator::sonar_front) >= 0
						 && readSonar->getMeasure(Configurator::sonar_back ) >= 0
						 && readSonar->getMeasure(Configurator::sonar_right) >= 0
						 && readSonar->getMeasure(Configurator::sonar_left) >= 0 ){

					NewMessage( MSG_FROM_SONAR, SONAR_DATA );
					
					if ( meas_progress == 1 ){
						AddBrianData( "SonarFront", readSonar->getMeasure(Configurator::sonar_front), 1.0 );
						AddBrianData( "SonarBack", readSonar->getMeasure(Configurator::sonar_back), 1.0 );
					}

					if ( meas_progress == 2 ){
						AddBrianData( "SonarRight", readSonar->getMeasure(Configurator::sonar_right), 1.0 );
						AddBrianData( "SonarLeft", readSonar->getMeasure(Configurator::sonar_left), 1.0 );
					}
					
					if ( meas_progress == 3 )
						meas_progress = 0;

					if ( !m_bInit ){
						settimer(0);
						m_bInit = true;
					}
				
					if ( elapsed(0) > Configurator::sonar_delay ){		// default 1 second
						AddBrianData( "SonarLeftOld", m_lBuffer.front(), 1.0 );
						AddBrianData( "SonarRightOld", m_rBuffer.front(), 1.0 );

						m_rBuffer.pop_front();
						m_lBuffer.pop_front();
					}

					m_rBuffer.push_back( readSonar->getMeasure( Configurator::sonar_right ) );
					m_lBuffer.push_back( readSonar->getMeasure( Configurator::sonar_left ) );


					for ( int i = 0; i < m_rBuffer.size(); i++ ){
						lLog << setw(4) << m_lBuffer[i] << ' ';
						rLog << setw(4) << m_rBuffer[i] << ' ';
					}

					lLog << endl;
					rLog << endl;

					CloseCurrentMessage();
					SendAllStringMessages();
					}
				}
			}
		} else {
			
		} 
	} while(line_readed<=max_line_err);
}

void SonarExpert::RunClose(){
	SetAllLeds(0);
	if ( readSonar )
		readSonar->sendStop();
		
	//FileLogger::writeLine( "Chiusura sonar" );
	
	if ( readSonar )
		delete readSonar;
		
	readSonar=NULL;
	
	//FileLogger::close();
	
	printf("\nSonarExpert closed\n");
}

void SonarExpert::SetAllLeds( char c ){
	m_ledValue.cValue = c;
	readSonar->sendStringCommand(&m_ledValue.cValue, 1);
}

void SonarExpert::SetGreenValue(bool b){
	m_ledValue.sValue.green = b ? 1 : 0;
	readSonar->sendStringCommand(&m_ledValue.cValue, 1);
}

void SonarExpert::SetRedValue(char c){
	m_ledValue.sValue.red = c;
	readSonar->sendStringCommand(&m_ledValue.cValue, 1);
}

void SonarExpert::SetYellowValue(char c){
	m_ledValue.sValue.yellow = c;
	readSonar->sendStringCommand(&m_ledValue.cValue, 1);
}

void SonarExpert::AddBrianData(char* name,float value,float rel){	
	if( fabs(value) < 0.001 ) //evito che vengano mandati valori in notazione scientifica
		value = 0;

	if( fabs(rel) < 0.001 )
		value = 0;
	
	mCurrentMessage << "<D>" << name << ' ' << value << ' ' << rel << " </D>\n";
}

long SonarExpert::elapsed(int i)
{ 
	struct timeval now;
	long elapsed_usec = 0;

	gettimeofday(&now, NULL);

	if (now.tv_usec > 1000000) {
		now.tv_usec -= 1000000;
		++now.tv_sec;
	}

	if (now.tv_sec > last_render[i].tv_sec)
		elapsed_usec = ((now.tv_sec - last_render[i].tv_sec) * 1000000);

	if (now.tv_usec > last_render[i].tv_usec)
		elapsed_usec += now.tv_usec - last_render[i].tv_usec;
	else
		elapsed_usec -= last_render[i].tv_usec - now.tv_usec;


	return elapsed_usec ;
}

void SonarExpert::settimer(int i){
	gettimeofday(&(last_render[i]),NULL);
}
