#include "SurfModule.h"

const CvScalar SURFModule::colors[] = { CV_RGB(255,0,0), CV_RGB(0,255,0), CV_RGB(0,0,255),
							            CV_RGB(255,255,0), CV_RGB(255,0,255), CV_RGB(0,255,255),
										CV_RGB(0,0,0), CV_RGB(255,255,255) };

int hdims = 16;
float hranges_arr[] = {0,180};
float* hranges = hranges_arr;

SURFModule::SURFModule( int w, int h ) : mHeight(h), mWidth(w), logger( log4cpp::Category::getInstance("SURF")) {
	mNModels = 0;
	mNumMatched = 0;
	mInit = false;
	
	pFileApp = new log4cpp::FileAppender( "vision.surf.file", "log/SurfLog.txt", false );
	pCoutApp = new log4cpp::OstreamAppender( "vision.surf.cout", &cout );

	pFileApp->setLayout( new log4cpp::BasicLayout() );
	pCoutApp->setLayout( new log4cpp::SimpleLayout() );
	logger.setAppender(pFileApp);
	logger.setAppender(pCoutApp);

	//pFileApp->setThreshold( log4cpp::Priority::NOTSET );
	pCoutApp->setThreshold(log4cpp::Priority::INFO );

	mStorage = cvCreateMemStorage(0);
	mLoopStorage = cvCreateMemStorage(0);

	curFrame = cvCreateImage( cvSize(w,h), 8, 3 );
	curGray = cvCreateImage( cvSize(w,h), 8, 1 );

	curHSV = cvCreateImage( cvSize(w,h), 8, 3 );
	curHUE = cvCreateImage( cvSize(w,h), 8, 1 );
	curValue = cvCreateImage( cvSize(w,h), 8, 1 );
	curBackProject = cvCreateImage( cvSize(w,h), 8, 1 );

	histImage = cvCreateImage( cvSize(320,200),8,3);

	cvZero(curBackProject);
	cvNot( curBackProject, curBackProject );

	//Homography matrix
	_hMatrix = cvMat( 3, 3, CV_64F, hMatrix );
	
	//Approssimazione a rettangolo
	mApproxedSquare = (CvContour*)cvCreateSeq( CV_SEQ_POLYGON, sizeof(CvContour), sizeof(CvPoint), mStorage );
	mSquareModel = (CvContour*)cvCreateSeq( CV_SEQ_POLYGON, sizeof(CvContour), sizeof(CvPoint), mStorage );

	mHist = cvCreateHist( 1, &hdims, CV_HIST_ARRAY, &hranges, 1 );

	// Rettangolo...
	squareModelPoints[0] = cvPoint(0,0);
	squareModelPoints[1] = cvPoint(1,0);
	squareModelPoints[2] = cvPoint(1,1);
	squareModelPoints[3] = cvPoint(0,1);
	
	cvSeqPush( (CvSeq*)mSquareModel, &squareModelPoints[0] );
	cvSeqPush( (CvSeq*)mSquareModel, &squareModelPoints[1] );
	cvSeqPush( (CvSeq*)mSquareModel, &squareModelPoints[2] );
	cvSeqPush( (CvSeq*)mSquareModel, &squareModelPoints[3] );
	
	cvNamedWindow( "hist" );
	cvNamedWindow( "back" );
}

SURFModule::~SURFModule(){
	cvReleaseImage( &curFrame );
	cvReleaseImage( &curHUE );
	cvReleaseImage( &curValue );

	for ( ModelsVector::iterator it = mModels.begin(); it != mModels.end(); it++ ){
		logger.info( "Model %s deleted\n", (*it)->file.c_str() );
		delete *it;
	}
		
	cvDestroyAllWindows();
	
	cvReleaseMemStorage( &mStorage );
	cvReleaseMemStorage( &mLoopStorage );
}


void SURFModule::process( IplImage *pFrame ){
	cvCopy( pFrame, curFrame, NULL );
	int mIndex = 0;
	
	createImages();
	//cvEqualizeHist(curGray,curGray);
	
	// *************************************** ESTRAZIONE SURF ************************************************** //
	CvSeq  *camKeypoints, *camDescriptors;		//keypoints e descrittori dell'immagine di spykee
	try {
		cvExtractSURF( curGray, curBackProject, &camKeypoints, &camDescriptors, mLoopStorage, cvSURFParams(SURF_THR,1) );
	} catch ( cv::Exception &e ){
		//deallocateHelper( camDescriptors, camKeypoints, mLoopStorage );
		cerr << "Error" << endl;
		cvClearMemStorage(mLoopStorage);
		return;
	}
	
	if ( camDescriptors->total <= 0 ){
		deallocateHelper( camDescriptors, camKeypoints, mLoopStorage );
		return;
	}

	int length = (int)(camDescriptors->elem_size/sizeof(float));
	cv::Mat camDescMatrix( camDescriptors->total, length, CV_32F );

	CvSeqReader reader;
	float *ptr = camDescMatrix.ptr<float>(0);
	cvStartReadSeq( camDescriptors, &reader );
	
	for(int i = 0; i < camDescriptors->total; i++ ){
		const float *pDescriptor = (const float*)reader.ptr;
		CV_NEXT_SEQ_ELEM( reader.seq->elem_size, reader );
		memcpy(ptr, pDescriptor, length*sizeof(float));
		ptr += length;
	}

	// ************************************************************************************************** //


	for ( mIndex = 0; mIndex < mModels.size(); mIndex++ ){
		ModelData *pModel = mModels.at(mIndex);

		CvPoint src_corners[4] = {{0,0}, {pModel->imgSize.width,0}, {pModel->imgSize.width, pModel->imgSize.height}, {0, pModel->imgSize.height}};
		CvPoint dst_corners[4];
		
		if ( camDescriptors->total <= 0 )
			break;

		// Inizializzazione KNN
		cv::flann::Index flann_index( camDescMatrix, cv::flann::KDTreeIndexParams(4) );
		flann_index.knnSearch( *(pModel->objDescriptors), *(pModel->indices), *(pModel->distances), 2, cv::flann::SearchParams(64) );

		/*************************************** Confronto SURF modello - SURF immagine -> KNN **********************/
		int *indicesPtr = pModel->indices->ptr<int>(0);
		float *distPtr = pModel->distances->ptr<float>(0);
		int count = 0;
	
		pModel->pt1.resize( pModel->indices->rows );
		pModel->pt2.resize( pModel->indices->rows );
		
		for ( int i = 0; i < pModel->indices->rows; i++ ){			
			CvSURFPoint *spt1 = ((CvSURFPoint*)cvGetSeqElem( pModel->objKeypoints, i ) );
			CvSURFPoint *spt2 = ((CvSURFPoint*)cvGetSeqElem( camKeypoints, indicesPtr[2*i] ));
				
			if ( distPtr[2*i] < 0.6*distPtr[2*i + 1] ){
				pModel->pt1[count] = spt1->pt;
				pModel->pt2[count++] = spt2->pt;
				
				cvCircle( curFrame, ::cvPointFrom32f(spt2->pt), cvRound(spt2->size*1.2/9.*2), colors[mIndex], 1 );
			}
		}

		/*************************************************************************************************************/

		/*********************************************** Se > 4 features -> homography *******************************/
		// Se < 4 carsh!!!! //

		if ( count >= 4 ){
			CvMat _pt1 = cvMat( 1, count, CV_32FC2, &(pModel->pt1[0]) );
			CvMat _pt2 = cvMat( 1, count, CV_32FC2,  &(pModel->pt2[0]) );
			
			if ( cvFindHomography( &_pt1, &_pt2, &_hMatrix, CV_RANSAC, 5 ) ){				
				for( int i = 0; i < 4; i++ ){
					double x = src_corners[i].x, y = src_corners[i].y;
					double Z = 1./(hMatrix[6]*x + hMatrix[7]*y + hMatrix[8]);		//scala ?
					double X = (hMatrix[0]*x + hMatrix[1]*y + hMatrix[2])*Z;
					double Y = (hMatrix[3]*x + hMatrix[4]*y + hMatrix[5])*Z;
					dst_corners[i] = cvPoint(cvRound(X), cvRound(Y));
				}
				
				cvSeqPush( (CvSeq*)mApproxedSquare, &dst_corners[0] );
				cvSeqPush( (CvSeq*)mApproxedSquare, &dst_corners[1] );
				cvSeqPush( (CvSeq*)mApproxedSquare, &dst_corners[2] );
				cvSeqPush( (CvSeq*)mApproxedSquare, &dst_corners[3] );
				
				CvSeq *res = cvApproxPoly( mApproxedSquare, mApproxedSquare->header_size, mLoopStorage, CV_POLY_APPROX_DP, cvContourPerimeter(mApproxedSquare)*0.02 );

				// Se assomiglia a un rettangolo....
				if ( ( cvMatchShapes( res, mSquareModel, CV_CONTOURS_MATCH_I1 ) < 0.1 ) &&
							 res->total == 4 && cvCheckContourConvexity(res) ){

					CvRect rc = points2Rect(dst_corners);
					cvDrawRect( curFrame, cvPoint(rc.x,rc.y), cvPoint(rc.x+rc.width,rc.y+rc.height), colors[mIndex], 3 );

					if ( mNumMatched < 3 )
						mNumMatched++;

					if ( !mInit && isDetected() ){
						try {
							cvSetImageROI( curHUE, rc );
							cvCalcHist( &curHUE, mHist, NULL, NULL );
							cvResetImageROI(curHUE);
							mInit = true;
						} catch ( cv::Exception &e ){
							cout << "OpenCV EXCEPTION: " << e.err << endl;
							break;
						}
					}

					float max_val = 0.0	;
					cvGetMinMaxHistValue( mHist, 0, &max_val, 0, 0 );
					cvConvertScale( mHist->bins, mHist->bins, max_val ? 255.0 / max_val : 0.0, 0 );

					// se è la prima volta che becco il marker creo anche l'istogramma del colore,
					// altrimenti uso quello del modello corrente
					if ( mInit ){
						cvCalcBackProject( &curHUE, curBackProject, mHist );
						drawHist( mHist, histImage );
					} else {
						cvCalcBackProject( &curHUE, curBackProject, pModel->pHist );
						drawHist( pModel->pHist, histImage );
					}

					// calcolo il centro del rettangolo che contiene il marker, in px
					if ( isDetected() )
						mCenter = cvPoint( rc.x + (rc.width/2) , rc.y + (rc.height/2) );

					cvShowImage( "hist", histImage );

					// pulisco i vector ---> poco efficente
					pModel->pt1.clear();
					pModel->pt2.clear();
					cvClearSeq( (CvSeq*)mApproxedSquare );
					break;
				}

				cvClearSeq( (CvSeq*)mApproxedSquare );

			} else {  //cvFindHomography 
				if ( mNumMatched > 0 )
					mNumMatched--;
				
				if ( mNumMatched == 0 ){
					cvZero(curBackProject);
					cvNot(curBackProject,curBackProject);
					mInit = false;
				}
			}

		} else  { // count >= 4
			if ( mNumMatched > 0 )
				mNumMatched--;

			if ( mNumMatched == 0 ){
				cvZero(curBackProject);
				cvNot(curBackProject,curBackProject);
				mInit = false;
			}
		}

		/****************************************************************************************************/

	} // Ciclo modelli

	cvShowImage( "back", curBackProject );

	//svuoto le pile e pulisco la memoria
	deallocateHelper( camDescriptors, camKeypoints, mLoopStorage );
}

void SURFModule::deallocateHelper( CvSeq *camDesc, CvSeq *camKeypts, CvMemStorage *pLoopStorage ){
	cvClearSeq(camDesc);
	cvClearSeq(camKeypts);
	cvClearMemStorage(pLoopStorage);
}


// Utility: calcolo immagini per ogni loop, sovrascrivo alle precedenti
void SURFModule::createImages(){
	cvCvtColor( curFrame, curGray, CV_BGR2GRAY );
	cvCvtColor( curFrame, curHSV, CV_BGR2HSV );
	cvSplit( curHSV, curHUE, NULL, NULL, NULL );
}

CvRect SURFModule::points2Rect(CvPoint pts[4]){
	int x1 = pts[0].x < pts[3].x ? pts[0].x : pts[3].x;
	int x2 = pts[1].x > pts[2].x ? pts[1].x : pts[2].x;
	int y1 = pts[0].y < pts[1].y ? pts[0].y : pts[1].y;
	int y2 = pts[2].y < pts[3].y ? pts[2].y : pts[3].y;

	return cvRect( x1, y1, x2 - x1, y2 - y1 );
}

// Carica i modelli dell'immagine
IplImage *SURFModule::LoadModel( string file, CvSeq **objKeypoints, CvSeq **objDescriptors, CvSize &size ){
	IplImage *pImage = cvLoadImage( file.c_str(), 1 );
	Ptr<IplImage> pGray = cvLoadImage( file.c_str(), 0 );
	
	cvCvtColor( pImage, pGray, CV_BGR2GRAY );
	
	size.height = pImage->height;
	size.width = pImage->width;
	
	cvExtractSURF( pGray, NULL, objKeypoints, objDescriptors, mStorage, cvSURFParams(SURF_THR,1) );

	for ( int i = 0; i < (*objKeypoints)->total; i++ ){
		CvSURFPoint *r = ((CvSURFPoint*)cvGetSeqElem( *objKeypoints, i ));
		CvPoint p = cvPointFrom32f(r->pt);
		cvCircle(pImage, p, cvRound(r->size*1.2/9.*2), colors[mNModels] );
	}

	return pImage;
}

void SURFModule::AddModel( string file ){
	ModelData *data = new ModelData;
	CvSeq *objDescriptors;
		
	data->pImage = LoadModel( file, &data->objKeypoints, &objDescriptors, data->imgSize );
	int length = (int)(objDescriptors->elem_size/sizeof(float));
	
	mNModels++;

	data->file = file;
	data->area = data->imgSize.width * data->imgSize.height;	
	data->objDescriptors = new cv::Mat( objDescriptors->total, length, CV_32F );
	data->indices = new cv::Mat( objDescriptors->total, 2, CV_32S );
	data->distances = new cv::Mat( objDescriptors->total, 2, CV_32F );
	
	data->pHist = cvCreateHist( 1, &hdims, CV_HIST_ARRAY, &hranges, 1 );
	
	Ptr<IplImage> hsv = cvCreateImage( cvGetSize(data->pImage), 8, 3 );
	IplImage *hue = cvCreateImage( cvGetSize(data->pImage), 8, 1 );
	
	cvCvtColor( data->pImage, hsv, CV_BGR2HSV );
	cvSplit( hsv, hue, NULL, NULL, NULL );
	cvCalcHist( &hue, data->pHist, NULL, NULL );
	cvReleaseImage(&hue);
	
	float max_val = 0.0	;
	cvGetMinMaxHistValue( data->pHist, 0, &max_val, 0, 0 );
	cvConvertScale( data->pHist->bins, data->pHist->bins, max_val ? 255. / max_val : 0., 0 );

	/* Copy model descriptors in a Mat class because of flann */
    CvSeqReader reader;
	float *ptr = data->objDescriptors->ptr<float>(0);
	
    cvStartReadSeq( objDescriptors, &reader );
    
    for(int i = 0; i < objDescriptors->total; i++ ) {
        const float *pDescriptor = (const float*)reader.ptr;
        CV_NEXT_SEQ_ELEM( reader.seq->elem_size, reader );
        memcpy(ptr, pDescriptor, length*sizeof(float));
        ptr += length;
	}
    
    cvClearSeq( objDescriptors );
    
	mModels.push_back( data );
	
	//printf( "Modello: %s\n\t\n\tKeypoints: %d\n\tSize: %dx%d\n", file.c_str(), data->objKeypoints->total, data->imgSize.width, data->imgSize.height );
	logger.info( "Modello:%s", file.c_str() );
	logger.info( "Keypoints: %d", data->objKeypoints->total );
	logger.info( "Size: %dx%d", data->imgSize.width, data->imgSize.height );
	logger.info("");
}

//Utility: disegna l'istogramma del colore
void SURFModule::drawHist( CvHistogram *hist, IplImage *histimg ){
	cvZero( histimg );
	int bin_w = histimg->width / hdims;

	for( int i = 0; i < hdims; i++ ) {
		int val = cvRound( cvGetReal1D(hist->bins,i)*histimg->height/255 );
		CvScalar color = hsv2rgb(i*180.f/hdims);
		cvRectangle( histimg, cvPoint(i*bin_w,histimg->height), cvPoint((i+1)*bin_w,histimg->height - val),
		color, -1, 8, 0 );
	}
}

//Struttura dati per i modelli
ModelData::ModelData(){
	pImage = NULL;
	objKeypoints = NULL;
	objDescriptors = NULL;
	indices = NULL;
	distances = NULL;
	imgSize.height = imgSize.width = 0;
	area = 0.0;
}

ModelData::ModelData( string _file, IplImage *image, CvSeq *objKey, Mat *objDesc, Mat *ind, Mat *dist, CvSize s, int a ){
	file = _file;
	pImage = image;
	objKeypoints = objKey;
	objDescriptors = objDesc;
	indices = ind;
	distances = dist;
	imgSize = s;
	area = a;
}

ModelData::~ModelData(){
	if ( pImage != NULL )
		cvReleaseImage( &pImage );
		
	cvClearSeq( objKeypoints );
	
	if ( objDescriptors )
		delete objDescriptors;
	
	if ( indices )
		delete indices;
		
	if ( distances )
		delete distances;
		
	if ( pHist )
		cvReleaseHist( &pHist );
}

