#ifndef SURF_MODULE_H
#define SURF_MODULE_H

#include <opencv/cv.h>
#include <opencv/highgui.h>

#include <ctime>
#include <stdlib.h>
#include <stdio.h>
#include <string>

#include <log4cpp/Category.hh>
#include <log4cpp/BasicLayout.hh>
#include <log4cpp/SimpleLayout.hh>
#include <log4cpp/FileAppender.hh>
#include <log4cpp/OstreamAppender.hh>

using namespace std;


#ifdef _WIN32
#ifdef _DEBUG
#pragma comment(lib,"cv200d.lib")
#pragma comment(lib,"cxcore200d.lib")
#pragma comment(lib,"highgui200d.lib")
#else
#pragma comment(lib,"cv200.lib")
#pragma comment(lib,"cxcore200.lib")
#pragma comment(lib,"highgui200.lib")
#endif
#endif

#include <string>
#include <iostream>

using namespace std;
using namespace cv;

#define IMAGE_WIDTH		320
#define IMAGE_HEIGHT 	240

#define SURF_THR 250

#define RED 0
#define GREEN 1
#define BLUE 2
#define ORANGE 3
#define VIOLET 4
#define YELLOW 5
#define BLACK 6
#define WHITE 7

class ModelData {
public:
	ModelData();
	ModelData( string _file, IplImage *image, CvSeq *objKey, Mat *objDesc, Mat *ind, Mat *dist, CvSize s, int a );
	~ModelData();
		
	vector<CvPoint2D32f> pt1;
	vector<CvPoint2D32f> pt2;
		
	string file;				// file
	IplImage *pImage;			// immagine
	CvSeq *objKeypoints;		// features surf
	Mat *objDescriptors;		// descrittori features surf
	Mat *indices;
	Mat *distances;
	CvSize imgSize;				// dimensioni dell'immagine
	int area;
	CvHistogram *pHist;			// istogramma del colore
};

typedef vector<ModelData*> ModelsVector;

class SURFModule {
public:
	SURFModule( int w, int h );
	~SURFModule();
	
	void process(IplImage *pFrame);
	void AddModel( string file );
	
	static const CvScalar colors[];

	IplImage *getFrame() const { return curFrame; }

	ModelsVector mModels;

	static void drawHist( CvHistogram *hist, IplImage *histimg );
	bool isDetected() const { return mNumMatched >= 1; }
	CvPoint getCenter() const { return mCenter; }

protected:
	IplImage *LoadModel( string file, CvSeq **objKeypoints, CvSeq **objDescriptors, CvSize &size );
	void createImages();
	CvRect points2Rect( CvPoint pts[4] );

	void deallocateHelper( CvSeq *camDesc, CvSeq *camKeypts, CvMemStorage *pLoopStorage );

	CvMemStorage *mStorage;
	CvMemStorage *mLoopStorage;
	int mNModels;
	
	int mNumMatched;
	CvPoint mCenter;

	CvSURFParams SURFParams;
	cv::flann::KDTreeIndexParams knnTreeParams;
	cv::flann::SearchParams searchParams;
	
	double hMatrix[9];
	CvMat _hMatrix;
			 
	CvPoint src_corners[4];
    CvPoint dst_corners[4];
	
	CvPoint squareModelPoints[4];
	CvContour *mSquareModel;
	CvContour *mApproxedSquare;
	Ptr<CvHistogram> mHist;
		
	IplImage *curFrame;
	Ptr<IplImage> curGray;
	Ptr<IplImage> curHSV;
	IplImage *curHUE;
	IplImage *curValue;
	Ptr<IplImage> curBackProject;
	Ptr<IplImage> histImage;

	bool mInit;
	int mHeight, mWidth;

	//logger stuff
	log4cpp::FileAppender *pFileApp;
	log4cpp::OstreamAppender *pCoutApp;
	log4cpp::Category &logger;
};


inline CvScalar hsv2rgb( float hue ){
    int rgb[3], p, sector;
    static const int sector_data[][3]=
        {{0,2,1}, {1,2,0}, {1,0,2}, {2,0,1}, {2,1,0}, {0,1,2}};
    hue *= 0.033333333333333333333333333333333f;
    sector = cvFloor(hue);
    p = cvRound(255*(hue - sector));
    p ^= sector & 1 ? 255 : 0;

    rgb[sector_data[sector][0]] = 255;
    rgb[sector_data[sector][1]] = 0;
    rgb[sector_data[sector][2]] = p;

    return cvScalar(rgb[2], rgb[1], rgb[0],0);
}

inline void drawInfo(IplImage *img, const char *s, int pos ){
  static int counter = 0;
  static clock_t t;
  static float fps;
  char fps_text[50] = {0};
  CvFont font;

  cvInitFont(&font,CV_FONT_HERSHEY_COMPLEX_SMALL, 1,1,0);

  // Add fps figure (every 10 frames)
	if (counter > 10) {
	  fps = (10.0f/(clock()-t) * CLOCKS_PER_SEC);
	  t=clock();
	  counter = 0;
	}

	counter++;

	// Get the figure as a string
	sprintf(fps_text,"FPS: %.2f %s", fps, s );

	// Draw the string on the image
	cvPutText (img,fps_text,cvPoint(10,25), &font, cvScalar(255,255,0));
}
#endif
