#include "VisionExpert.h"

VisionExpert::VisionExpert(ModuleDispatch* kernel, int period, Spykee *s, SpykeeGui *g, DCDT_Mutex *mutex )
	: StringModuleMember(kernel, period, "VisionExpert"), m_logger( "/log/VisionLog.txt", true ){
	pSpykee = s;
	gui = g;
	pMutex = mutex;
	
	this->pSurf = new SURFModule(320,240);

	if ( pSurf == NULL ){
		cout << "Unable to initialize vision module! Memory?" << endl;
		this->Shutdown(1);
	}

	m_bRecordVideo = false;
	
	if ( pSpykee )
		pSpykee->startCamera();
	
	printf( "VisionExpert()\n" );
}

VisionExpert::~VisionExpert(){
	if ( pSurf )
		delete pSurf;

	printf( "\nVisionExpert::~VisionExpert()\n" );
}

void VisionExpert::RunInit(){
	printf( "VisionExpert::RunInit()\n" );

	pSurf->AddModel("./models/big.bmp");
	pSurf->AddModel("./models/big2.bmp");
	pSurf->AddModel("./models/small.bmp");
	pSurf->AddModel("./models/medium.bmp");

	m_bRecordVideo = false;
	mFramesCount = 0;
}

void VisionExpert::RunDuty(){
	int imageLen = 0;
	char imgName[50] = {0};

	if ( !pSpykee ){
		cout << "pSpykee == NULL" << endl;
		return;
	}

	if ( pMutex )
		pMutex->lock();

	unsigned char *jpegBuffer = pSpykee->getImage( imageLen );

	if ( jpegBuffer == NULL || imageLen <= 0 ){
		printf( "\nImage null!\n" );
		return;
	}

	CvMat jpgMat = cvMat( 1, imageLen, CV_8UC1, jpegBuffer );
	IplImage *pFrame =  cvDecodeImage( &jpgMat, 1 );
	IplImage *pResFrame = NULL;

	if ( true ){
		this->pSurf->process(pFrame);
		pResFrame = pSurf->getFrame();

		NewMessage( MSG_FROM_VISION, VISION_DATA );

		if ( pSurf->isDetected() ){
			CvPoint center = pSurf->getCenter();

			AddBrianData( "VisionPlayerDetected", 1.0, 1.0 );
			AddBrianData( "PlayerAngle", center.x, 1 );
			//AddBrianData( "PlayerAngle", ((center.x * 50.0) / 320.0) - 25, 1 );
			//cout << "Position: " << center.x << endl;

			cvLine( pResFrame, cvPoint(center.x,0), cvPoint(center.x,240), CV_RGB(255,255,255), 2 );
			cvLine( pResFrame, cvPoint(0,center.y), cvPoint(320,center.y), CV_RGB(255,255,255), 2 );
		} else {
			AddBrianData("VisionPlayerDetected", 0.0, 1.0 );
		}

		CloseCurrentMessage();
		SendAllStringMessages();
	}

	if ( pMutex )
		pMutex->unlock();

	if ( m_bRecordVideo )
		drawInfo(pResFrame, "Rec", 0 );
	else
		drawInfo(pResFrame, "", 0 );

	if ( gui )
		gui->showImage(pResFrame);
	else
		cvShowImage("img", pResFrame);

	char c = cvWaitKey(10) & 0xFF;

	switch(c){
		case 27:   //ESC
			this->Shutdown(1);
			break;
		case 's':
			sprintf( imgName, "./pics/pic %d.bmp\0", mFramesCount++ );
			cvSaveImage( imgName, pFrame );
			cout << "Screenshot taken: " << imgName << endl;
			break;
		case 'r':
			if ( !m_bRecordVideo ){
				m_bRecordVideo = true;
				m_videoWriter = cvCreateVideoWriter( "./pics/video.avi", CV_FOURCC('t', 'h', 'e', 'o'),  10, cvSize(320,240), 1 );
				cout << "Recording video..." << endl;
			} else {
				cvReleaseVideoWriter(&m_videoWriter);
				m_bRecordVideo = false;
				cout << "Stop recording." << endl;
			}
			break;
	}

	if ( m_bRecordVideo )
		cvWriteFrame(m_videoWriter, pResFrame);

	cvReleaseImage(&pFrame);
}

void VisionExpert::RunClose(){
	printf( "\nVisionExpert::RunClose()\n" );
	
	if ( m_bRecordVideo )
		cvReleaseVideoWriter( &m_videoWriter );
}

void VisionExpert::AddBrianData(char* name,float value,float rel){
	if(fabs(value) < 0.001) //evito che vengano mandati valori in notazione scientifica
		value = 0;

	if(fabs(rel) < 0.001 )
		rel = 0;
	
	mCurrentMessage << "<D>" << name << " " << value << " " << rel << " </D>\n";
}
