#ifndef VisionExpert_h
#define VisionExpert_h

#include "StringModuleMember.h"
#include "SurfModule.h"

#include <sys/time.h>
#include <stdlib.h> 
#include <stdio.h>
#include <const.h>
#include <msg.h>

#include <sys/time.h>

#include "Spykee.h"
#include "SpykeeGui.h"
#include "FileLogger.h"

#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <opencv/cxcore.h>

using namespace std;
using namespace cv;

#define IMAGE_WIDTH		320
#define IMAGE_HEIGHT 	240

#define SURF_THR 225

#define RED 0
#define GREEN 1
#define BLUE 2
#define ORANGE 3
#define VIOLET 4
#define YELLOW 5
#define BLACK 6
#define WHITE 7

static const CvScalar colors[] = { CV_RGB(255,0,0), CV_RGB(0,255,0), CV_RGB(0,0,255),
								   CV_RGB(255,255,0), CV_RGB(255,0,255), CV_RGB(0,255,255),
								   CV_RGB(0,0,0), CV_RGB(255,255,255) };
								   
								   
class VisionExpert : public StringModuleMember {
	class ModelData {
		public:
		ModelData();
		ModelData( string _file, IplImage *image, CvSeq *objKey, Mat *objDesc, Mat *ind, Mat *dist, CvSize s, double a );
		~ModelData();
		
		vector<CvPoint2D32f> pt1;
		vector<CvPoint2D32f> pt2;
		
		string file;
		IplImage *pImage;
		CvSeq *objKeypoints;
		Mat *objDescriptors;
		Mat *indices;
		Mat *distances;
		CvSize imgSize;
		double area;
	};
	
	typedef vector<VisionExpert::ModelData*> ModelsVector;
	
	public:
	VisionExpert(ModuleDispatch* kernel, int period, Spykee *s, SpykeeGui *g, DCDT_Mutex *mutex );
	~VisionExpert();
	
	void RunInit();
	void RunDuty();
	void RunClose();
	
	protected:
	void AddBrianData(char* name,float value,float rel);
	
	Spykee *pSpykee;
	
	bool m_bRecordVideo;
	CvVideoWriter *m_videoWriter;
	CvCapture *pCap;
	unsigned int mFramesCount;
	
	IplImage *videoImage;

	private:
	SpykeeGui *gui;
	DCDT_Mutex *pMutex;
	FileLogger m_logger;
	SURFModule *pSurf;
};

#endif
