#include "WiimExpert.h"
#include <msg.h>

#include <iostream>
#include <math.h>
#include <string>
#include <stdlib.h>


#include "Configurator.h"

#define PI 3.14159265358979323846

using namespace std;

#define FPS_TIMER 		0
#define CLOCK_TIMER 	1
#define GAME_TIMER 		5
#define GAMESTART_TIMER 6	

// la parola extern non dichiara un nuovo identificatore, 
// ma dice "La variabile intera a e` dichiarata da qualche altra parte, 
// lascia solo lo spazio per risolvere il riferimento"
extern void wiimlex_memory(const char* src, WiimExpert* pWiim);
extern int MessageLineNumWiim;

extern struct wiimote_t* wm0;

WiimExpert::WiimExpert(ModuleDispatch* kernel,int period, Spykee* s) : StringModuleMember(kernel,period,"WiimExpert"){
  pSpykee = s;
  printf("\nWiimExpert built");
  fflush(stdout);
};

WiimExpert::~WiimExpert()
{
  fclose(fp);
  fclose(f);
  printf("\nWiimExpert destroyed");
  fflush(stdout);
};

void WiimExpert::RunInit(){
	AddMessageRequest( MSG_FROM_BRIAN, LOCAL );

	int i=0;
	
	m_bAuto = Configurator::initAuto;
	
	charged=0;
	Hit=0;
	fps=0;
	random_sign=2;
	
	settimer(0);
	settimer(1);
	settimer(2);
	settimer(3);
	
	center.x=0;
	center.y=0;
	
	robotpoint=0;
	humanpoint=0;
	pointrobotflag=0;
	
	srand(time(NULL));
	
	for(i=0;i<IRCENTERFILTER;i++){
		ir_centerarray_x[i].v=0;
		ir_centerarray_x[i].w=0;
		ir_centerarray_y[i].v=0;
		ir_centerarray_y[i].w=0;
	}
	
	wiim = new Wiim(); //inizializza (connessione al wiimote)
	
	struct timeval now;
	gettimeofday(&now,NULL);
    base_time=now.tv_sec;

	playing = false;
	clock = 0;
	cameraLightOn = false;
	settimer(GAMESTART_TIMER);
  	
	if (pSpykee != NULL)
		pSpykee->turnOffCameraLight();

  	printf("\nWiimExpert initialized");
  	fflush(stdout);
};

/* ***********************************
 
   DoYourDuty()

   *********************************** */
void WiimExpert::RunDuty(){
	ppoint ir_center;
	float ir_centerx,ir_centery;
	int ir_centervalid;
	weightfloat newdata;
	weightfloat ir_newcenterx,ir_newcentery;
	float ir_center_distance=0;
	ostringstream oss;
	
	// --- blocco d'emergenza ---
	if ( wiim->getPButtonTwo() ){
		NewMessage(MSG_TO_MOTION,WIIMOTE_DATA);
		mCurrentMessage << "<C>TanSpeed 0.0</C>\n" << "<C>RotSpeed 0.0</C>\n";
		CloseCurrentMessage();
		SendAllStringMessages();

		int events = wiim->run();
		if( wm0==NULL || !events )
			return;
	}
	
	// --- End game --- //
	if ( wiim->getPButtonHome() ){
		cout << "Shutting down" << endl;

		NewMessage(MSG_TO_MOTION, WIIMOTE_DATA);
		mCurrentMessage << "<C>TanSpeed 0.0 </C>\n" << "<C>RotSpeed 0.0</C>\n";
		CloseCurrentMessage();
		SendAllStringMessages();

		NewMessage( MSG_TO_SONAR, LED_DATA );
		mCurrentMessage << "Red " << 0 << " Green " << 0 << "\n";
		CloseCurrentMessage();
		SendAllStringMessages();
		this->Shutdown(1);

		return;
	}

	/*
	 * Passo in automatico (gioco)
	 */
	if ( !m_bAuto && wiim->getPButtonPlus() ){
		m_bAuto = true;
		settimer(GAMESTART_TIMER);
		settimer(CLOCK_TIMER);
		settimer(GAME_TIMER);
		settimer(FPS_TIMER);
	}

	/*
	 * Controllo manuale
	 */
	if ( m_bAuto && wiim->getPButtonMinus() ){
		m_bAuto = false;
		playing = false;
		cout << "Manual mode\nBrian deactivated" << endl;

		// Reset brian
		NewMessage( MSG_TO_SONAR, LED_DATA );
		mCurrentMessage << "Red " << 4 << "\nGreen " << 8 << "Yellow " << 255 << "\n";
		CloseCurrentMessage();

		NewMessage( MSG_FROM_WIIMOTE, WIIMOTE_DATA );
		AddBrianData( "GameStatus", 0, 1 );
		CloseCurrentMessage();

		SendAllStringMessages();
	}

	// --- Controllo manuale --- //
	if ( !m_bAuto ){
		NewMessage( MSG_FROM_WIIMOTE, WIIMOTE_DATA );
		AddBrianData( "GameStart", 0, 1 );
		CloseCurrentMessage();
		SendAllStringMessages();

		NewMessage( MSG_TO_MOTION, WIIMOTE_DATA );

		if ( wiim->getPButtonUp() )
			mCurrentMessage << "<C>TanSpeed 75.0</C>\n";
		else if ( wiim->getPButtonDown() )
			mCurrentMessage << "<C>TanSpeed -75.0</C>\n";
		else if ( wiim->getPButtonRight() )
			mCurrentMessage << "<C>RotSpeed 50.0</C>\n";
		else if ( wiim->getPButtonLeft() )
			mCurrentMessage << "<C>RotSpeed -50.0</C>\n";
		else
			mCurrentMessage << "<C>TanSpeed 0.0</C>\n<C>RotSpeed 0.0</C>\n";

		CloseCurrentMessage();
		SendAllStringMessages();

		settimer(GAMESTART_TIMER);

		if((wm0==NULL) || (!wiim->run()))
			return;

		return;
	}

	if (elapsed(CLOCK_TIMER) > 1000000){		// Ogni secondo mando il clock
		clock = ( clock == 0 ? 1 : 0 );
		random_sign = rand() % 6;
		settimer(CLOCK_TIMER);
	}

	if ( m_bAuto && ((elapsed(GAMESTART_TIMER) > 5000000) && (playing == false)) ){
		printf("\n--- *** START PLAY *** ---");

		wiim->WeakRumble(1);
		playing = true;
		Hit = 0;
		charged=0;

		settimer(GAME_TIMER); // Timer per il tempo di durata del gioco

		NewMessage( MSG_FROM_WIIMOTE, WIIMOTE_DATA );
		AddBrianData( "GameStart", 1, 1 );
		CloseCurrentMessage();
		SendAllStringMessages();
	}

	int events = wiim->run();
	if((wm0==NULL) || (!events))
		return;
	
	ir_center=wiim->getcenter();
	
	if ( ir_center.valid == 0 ){
		newdata.w = 0;

		ir_newcenterx=getFilteredData(newdata,ir_centerarray_x,IRCENTERFILTER,IRCENTERATT,NULL);
		ir_newcentery=getFilteredData(newdata,ir_centerarray_y,IRCENTERFILTER,IRCENTERATT,NULL);

		if(ir_newcenterx.w==0 || ir_newcentery.w==0)
			ir_centervalid = 0;

	} else {
		newdata.w = 1;

		newdata.v = ir_center.x;
		ir_newcenterx=getFilteredData(newdata,ir_centerarray_x,IRCENTERFILTER,IRCENTERATT,NULL);

		newdata.v=ir_center.y;
		ir_newcentery=getFilteredData(newdata,ir_centerarray_y,IRCENTERFILTER,IRCENTERATT,NULL);

		ir_centervalid = 1;
	}

	ir_centerx=ir_newcenterx.v;
	ir_centery=ir_newcentery.v;

	if(ir_centervalid==1)
		ir_center_distance=sqrt(pow((ir_centerx-1024/2),2)+pow((ir_centery-768/2),2));
	
	// --- messaggi al log ---
	NewMessage(MSG_TO_LOG,WIIMOTE_DATA);
	if(ir_centervalid==1)
		mCurrentMessage << "valid 1 xcenter "<<ir_centerx<<" ycenter "<<ir_centery;
	else
		mCurrentMessage << "valid 0 xcenter 0 ycenter 0";

	CloseCurrentMessage();
	SendAllStringMessages();
	
	if( elapsed(FPS_TIMER) > 1000000 ){
		settimer(0);
		printf("\nWiimExpert (%d) fps:%d     ", m_bAuto, fps);
		printstatus();
		fps = 0;
	} else
		fps++;
	
	//carica
	wiim->SetCharge(charged);

	if (ir_centervalid){
		if(ir_center_distance<IRCENTERLIMIT){

			NewMessage(MSG_TO_SONAR, LED_DATA);
			mCurrentMessage << "Red " << charged << " Green " << 8 << "\n";
			CloseCurrentMessage();
			SendAllStringMessages();

			if( charged < 4.5 )
				charged += 0.1;

		} else {
			NewMessage(MSG_TO_SONAR, LED_DATA);
			mCurrentMessage << "Red " << charged << " Green " << 0 << "\n";
			CloseCurrentMessage();
			SendAllStringMessages();

			if(charged >=0)
				charged -= 0.02;
		}
	}else{
		NewMessage(MSG_TO_SONAR, LED_DATA);
		mCurrentMessage << "Red " << charged << " Green " << 0 << "\n";
		CloseCurrentMessage();
		SendAllStringMessages();

		if(charged >=0)
			charged -= 0.02;
	}
	
	if ( (charged > 4) && (ir_centervalid) && (ir_center_distance < IRCENTERLIMIT) ){
		if( wiim->getPButtonB() ){
			if( Hit == 0 ){
				Hit = 1;
				charged = 0;
				cout << "\n--- *** COLPITO! *** ---" << endl;
				humanpoint++;
				playing = false;

				if (pSpykee != NULL)
					pSpykee->soundAllarm();
					
				wiim->SetpointDiff(humanpoint-robotpoint);
				wiim->BlinkRight(1);
				wiim->StrongRumble(1);

				NewMessage( MSG_FROM_WIIMOTE, WIIMOTE_DATA );
				AddBrianData( "GameStatus", 0, 1 );
				AddBrianData( "GameStart", 0, 1 );
				CloseCurrentMessage();
				SendAllStringMessages();

				NewMessage(MSG_TO_MOTION, COMMAND_DATA);
				mCurrentMessage << "<C>TanSpeed " << 0.0 << "</C>\n";
				mCurrentMessage << "<C>RotSpeed " << 0.0 << "</C>\n";
				CloseCurrentMessage();
				SendAllStringMessages();

				settimer(GAMESTART_TIMER);
			}
		}
	}
	
	NewMessage( MSG_TO_SONAR, LED_DATA );
	mCurrentMessage << "Yellow " << humanpoint << "\n";
	CloseCurrentMessage();
	SendAllStringMessages();


	if ( humanpoint == 3 ){
		cout << "Vittoria! Hai colpito il robot per 3 volte!\n"
			 << "Premi \'+\' sul Wiimote per iniziare una nuova partita" << endl;

		playing = m_bAuto = false;
		humanpoint = robotpoint = 0;

		if ( pSpykee )
			pSpykee->soundAllarm();

		wiim->BlinkRight(1);
		wiim->StrongRumble(1);
		return;
	}

	if( wiim->getPButtonB() ){
		if( charged > 2 )
			wiim->WeakRumble(1);
			
		charged = 0;
	}
	
	//punto robot
/*	if ((elapsed(GAME_TIMER) > 60000000) && (Hit == 0))	{
		robotpoint++;
		printf("\n--- *** punto per il robot *** ---");
		playing = false;
		
	//  if (pSpykee != NULL)
	//		pSpykee->soundBomb();
			
		Hit = -1;
		wiim->SetpointDiff(humanpoint-robotpoint);
		wiim->BlinkLeft(1);
		wiim->StrongRumble(1);
		settimer(GAMESTART_TIMER);
	}
	*/
	
	if ( playing == true ){
		NewMessage( MSG_FROM_WIIMOTE, WIIMOTE_DATA );
		AddBrianData("Random",random_sign,1);
		AddBrianData("Hit", Hit, 1);
		AddBrianData("Clock", clock, 1);
	
		//messaggio distanza e direzione dal centro IR (per brian)
	
		if(ir_centervalid){
			if(ir_centerx<1024/2)
				AddBrianData("IRCenter",-1*ir_center_distance,1.0);
			else
				AddBrianData("IRCenter",+1*ir_center_distance,1.0);

			AddBrianData( "IRDist", wm0->ir.z, 1.0 );
		}
		CloseCurrentMessage();
		//invio tutti i msgs
		SendAllStringMessages();
	}
}


void WiimExpert::RunClose(){
  printf("\nWiimExpert closed");
  fflush(stdout);
};

void WiimExpert::AddBrianData(char* name,float value,float rel){
	float v;
	float r;
	
	v=value;
	r=rel;
	if(fabs(v)<0.001){ //evito che vengano mandati valori in notazione scientifica
		v=0;
	}
	if(fabs(r)<0.001){
		v=0;
	}
	
	mCurrentMessage<<"<D>"<<name<<" "<<v<<" "<<rel<<" </D>\n";
	return;
	
}

long WiimExpert::elapsed(int i){
	struct timeval now;
	long elapsed_usec = 0;

	gettimeofday(&now, NULL);

	if (now.tv_usec > 1000000) {
		now.tv_usec -= 1000000;
		++now.tv_sec;
	}

	if (now.tv_sec > last_render[i].tv_sec)
		elapsed_usec = ((now.tv_sec - last_render[i].tv_sec) * 1000000);

	if (now.tv_usec > last_render[i].tv_usec)
		elapsed_usec += now.tv_usec - last_render[i].tv_usec;
	else
		elapsed_usec -= last_render[i].tv_usec - now.tv_usec;


	return elapsed_usec ;

}

void WiimExpert::settimer(int i){
	gettimeofday(&(last_render[i]),NULL);
}

void::WiimExpert::printstatus(){
	unsigned char *buffer = NULL;
	float v = -1;

	if ( ReceiveLastMessage( MSG_FROM_BRIAN, buffer, false ) > 0 )
		strtovalue( (char*)buffer, "GameStatus ", &v );

	printf("carica: %f\t",charged); 

	if( Hit )
		printf("Colpito!\t");

	printf(" hum:%d rob:%d\t",humanpoint,robotpoint);

	switch ( (int)v ){
		case 0:
			printf("S:Search");
			break;
		case 1:
			printf("S:Escape");
			break;
		case 2:
			printf("S:Hide");
			break;
		default:
			printf("E:?" );
	}

	printf("\n");
}

void::WiimExpert::setRotDirection(int dir){
	//rotDirection = dir;
}
