#include <const.h>
#include <string>
#include <type.h>
#include <cmath>

#include "Spykee.h"

#define SPYKEE


#define BRIAN
#define WIIM
#define LOG
#define MOTOR

#define MOTOR_DEBUG
#define MOTOR_DETAIL
#define MOTOR_FORCE_DEBUG

#ifdef MOTOR_FORCE_DEBUG
#define MOTOR_DEBUG
#endif

#ifdef MOTOR
	#ifdef MOTOR_DETAIL
		#define MOTOR_DEBUG
	#endif
#endif

#define SONAR

#ifdef SPYKEE
	#define VISION
#endif

//#define PC_GUI

#ifdef BRIAN
#include "BrianExpert.h"
int brian_period = 1; //in microsecondi
#endif

#ifdef SONAR
#include "SonarExpert.h"
int sonar_period=0;
#endif 

#ifdef WIIM
#include "WiimExpert.h"
int wiim_period = 100;
#endif

#ifdef MOTOR
#include "MotorExpert.h"
int motor_period=150000;
#endif

#ifdef LOG
#include "LogExpert.h"
int log_period=1;
#endif

#ifdef VISION
#include "VisionExpert.h"
int Vision_period = 0;
#endif

#include "ControlExpert.h"
#include "SpykeeGui.h"
int gui_period = 0;


#include "Configurator.h"

// DCDT Configuration files
string dcdt_config_file("./config/dcdt.conf");
string default_hostname("hostname");
string spykee_config_file("./config/spykee.cfg");

void OnSignal( int s );
ModuleDispatch *pDispatch = NULL;

int main(int argc, char** argv){
	Spykee *s = NULL;

	if ( !Configurator::init(spykee_config_file, argc, argv) )
		cout << "Unable to load configuration file!" << endl;

#ifdef SPYKEE
	if ( !Configurator::simulation ){
		s = new Spykee( Configurator::spykee_user, Configurator::spykee_pass );
		s->unplug();
		s->startCamera();
		s->turnOffCameraLight();
	} else {
		s = NULL;
	}
#endif

  // Create the Dispatcher
  ModuleDispatch* Dispatch = new ModuleDispatch ("", basename(argv[0]));
  pDispatch = Dispatch;
  
  // Create the Agent list
  std::vector< StringModuleMember*> agents;

  // ----------- BRIAN ----------- //
#ifdef BRIAN
 	BrianExpert* pBrian = new BrianExpert (Dispatch,brian_period);
 	agents.push_back(pBrian);
 	cout << endl << "--- BrianExpert Created --- "<<endl; 
#endif
 
  // ----------- MOTOR ----------- //
#ifdef MOTOR
 	MotorExpert* pMotor = new MotorExpert (Dispatch, motor_period, s); 
 	agents.push_back(pMotor);
	cout << endl << "--- MotorExpert Created --- "<<endl; 
#endif

#ifdef LOG  
	LogExpert* pLog = new LogExpert (Dispatch, log_period);
	agents.push_back(pLog);
	cout << endl << "--- LogExpert Created --- "<<endl; 
#endif

#ifdef PC_GUI
	DCDT_Mutex mMutex;
	SpykeeGui *pSpykeeGui = new SpykeeGui( &mMutex, &argc, &argv );
	ControlExpert *pControl = new ControlExpert( Dispatch, 0, pSpykeeGui );
	agents.push_back(pControl);
	cout << endl << "--- ControlExpert Created --- " << endl;
#else
	SpykeeGui *pSpykeeGui = NULL;
#endif


	// -------------- VISION ----------------- //
#ifdef VISION
	if ( !Configurator::simulation ){
		#ifdef PC_GUI
			VisionExpert* pVision = new VisionExpert (Dispatch, Vision_period, s, pSpykeeGui, &mMutex );
		#else
			VisionExpert* pVision = new VisionExpert (Dispatch, Vision_period, s, NULL, NULL );
		#endif

		agents.push_back(pVision);
		cout << endl << "--- Vision Created --- "<<endl;
	}
#endif

  // ----------- WIIM ----------- //
#ifdef WIIM  MemFlag
	WiimExpert* pWiim = new WiimExpert (Dispatch, wiim_period, s );
	agents.push_back(pWiim);
	cout << endl << "--- WiimExpert Created --- "<<endl; 
#endif

  // ----------- SONAR ----------- //
#ifdef SONAR
 	SonarExpert* pSonar = new SonarExpert (Dispatch, "SonarExpert", sonar_period); 
 	agents.push_back(pSonar);
 	cout << endl << "--- SonarExpert Created --- "<<endl; 
#endif

 	for (std::vector < StringModuleMember * >::iterator i = agents.begin (); i != agents.end (); i++)
 		Dispatch->AddMember (*i);
  
 	cout << endl << "--- Community Created --- "<<endl;
  
 	for (std::vector < StringModuleMember * >::iterator i = agents.begin (); i != agents.end (); i++)
 		Dispatch->ActivateMember (*i);
  
	signal( SIGINT, OnSignal );
	signal( SIGTERM, OnSignal );

	cout << endl << "--- ''let's work, be proud...'' --- "<<endl;
	Dispatch->LetsWork ();
	
	signal( SIGINT, 0 );
	signal( SIGTERM, 0 );

	cout << endl << "--- Community Session Terminated --- "<<endl;

	return 0;
}

void OnSignal( int s ){
	switch(s){
		case SIGINT:
		case SIGTERM:
			cout << "\nCtrl-c detected -> Shutting down..." << endl;
			pDispatch->SetStatus(DCDT_TERMINATING);
		break;
	}

	signal(s,OnSignal);
}
